/*
  Author: Shane Neph & Scott Kuehn
  Date:   Thu Dec 17 20:01:34 PST 2009
*/
//
//    BEDOPS
//    Copyright (C) 2011-2020 Shane Neph, Scott Kuehn and Alex Reynolds
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License along
//    with this program; if not, write to the Free Software Foundation, Inc.,
//    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

#ifndef _BED_WINDOW_VISITOR_TYPES_HPP
#define _BED_WINDOW_VISITOR_TYPES_HPP

#include "bed/BedBaseVisitor.hpp"
#include "bed/EchoMapBedVisitor.hpp" // do not include other/EchoMapVisitor.hpp
#include "bed/EchoMapIntersectLengthVisitor.hpp"
#include "bed/OvrAggregateVisitor.hpp"
#include "bed/OvrUniqueVisitor.hpp"
#include "bed/OvrUniqueFractionVisitor.hpp"
#include "bed/WeightedAverageVisitor.hpp"
#include "NumericalVisitors.hpp"
#include "other/EchoVisitor.hpp"
#include "other/MultiVisitor.hpp"
#include "Visitors.hpp"

#endif // _BED_WINDOW_VISITOR_TYPES_HPP
