#!/bin/sh

rm -f /Library/Application\ Support/BEDOPS/libstdc++.6.dylib
rm -f /Library/Application\ Support/BEDOPS/libgcc_s.1.dylib
rm -f /usr/local/bin/bedextract
rm -f /usr/local/bin/bedmap
rm -f /usr/local/bin/bedops
rm -f /usr/local/bin/closest-features
rm -f /usr/local/bin/sort-bed
rm -f /usr/local/bin/starch
rm -f /usr/local/bin/starchcat
rm -f /usr/local/bin/unstarch
rm -f /usr/local/bin/starchcluster
rm -f /usr/local/bin/starchcluster.gnu_parallel
rm -f /usr/local/bin/bam2bed
rm -f /usr/local/bin/bam2starch
rm -f /usr/local/bin/bam2bed_sge
rm -f /usr/local/bin/bam2starch_sge
rm -f /usr/local/bin/bam2bed_gnuParallel
rm -f /usr/local/bin/bam2starch_gnuParallel
rm -f /usr/local/bin/gff2bed
rm -f /usr/local/bin/gff2starch
rm -f /usr/local/bin/gtf2bed
rm -f /usr/local/bin/gtf2starch
rm -f /usr/local/bin/psl2bed
rm -f /usr/local/bin/psl2starch
rm -f /usr/local/bin/sam2bed
rm -f /usr/local/bin/sam2starch
rm -f /usr/local/bin/vcf2bed
rm -f /usr/local/bin/vcf2starch
rm -f /usr/local/bin/wig2bed
rm -f /usr/local/bin/wig2bed_bin
rm -f /usr/local/bin/wig2starch
rm -f /usr/local/bin/convert2bed

exit 0


